// Brutal LZSS compressor

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define WINDOW 4096 // Window size
#define SHIFT 4 // How many bits for the count
#define MAXLEN 16 // Maximum length of referenced data

// Write a bit into a file with optional flush
void writebit(FILE *f,char bit,int flush)
{
    static unsigned char byte=0;
    static int bitnum=7;

    byte+=bit<<bitnum;

    bitnum--;
    if(bitnum<0)
    {
        fputc(byte,f);
        bitnum=7;
        byte=0;
    }

    if(flush && bitnum!=7)
        fputc(byte,f);
}

// Write a full byte into a file
void writebyte(FILE *f,unsigned char c,int flush)
{
    int n;
    for(n=0;n<7;n++)
        writebit(f,(c>>(7-n))&1,0);
    writebit(f,c&1,flush);
}

// (Brutally) find the longest match of bytes in the preceding window
int findmatch(unsigned char *data,int prev,int cur,int end,int *ref)
{
    int longest=0,i,j,k;
    *ref=0;

    for(i=prev;i<cur;i++) // Go through the window
    {
        for(j=cur,k=0;j<end && k+i<cur;j++,k++) // Compare to the current block
            if(data[i+k]!=data[j])
                break;

        if(k>longest)
        {
            longest=k;
            *ref=cur-i;
        }
    }

    return longest;
}

int main(int argc,char *argv[])
{
    int length;
    FILE *s,*d;
    unsigned char *dada;

    if(argc!=3)
    {
        printf("lzss source dest.lzss\n");
        return 1;
    }

    // Open dest and src
    s=fopen(argv[1],"rb");
    fseek(s,0,SEEK_END);
    length=ftell(s);
    fseek(s,0,SEEK_SET);

    d=fopen(argv[2],"wb");
    if(s==NULL || d==NULL)
        return 1;

    dada=malloc(length);
    fread(dada,1,length,s);
    fclose(s);

    int i,prev,longest,end,ref;

    // Ok, let's get started with the compression
    for(i=0;i<length;)
    {
        prev=i-WINDOW; // Start of search window in previous data
        if(prev<0)
            prev=0;

        end=i+MAXLEN; // Maximum index of the current block (+1)
        if(end>length)
            end=length;

        // Find the longest match in the data
        longest=findmatch(dada,prev,i,end,&ref);

        if(longest<2) // Too short match to compress
        {
            writebit(d,0,0);
            writebyte(d,dada[i],0);
            i++;
        }
        else // Reference to previous data
        {
            int word=0;

            word+=(ref-1)<<SHIFT;
            word+=longest-1;

            writebit(d,1,0);
            writebyte(d,word>>8,0);
            writebyte(d,word&0xff,0);
            i+=longest;
        }
    }

    // Terminate with bit 1 and two zero bytes and flush to full bytes
    writebit(d,1,0);
    writebyte(d,0,0);
    writebyte(d,0,1);

    fclose(d);

    s=fopen(argv[2],"rb");
    fseek(s,0,SEEK_END);
    length=ftell(s);
    fclose(s);

    d=fopen("kludge.tmp","wb");
    fputc((length&0xff00)>>8,d);
    fputc((length&0xff)>>0,d);
    fclose(d);

    return 0;
}
